import sys
import functools


def env_error_to_os_specific(func):
    @functools.wraps(func)
    def res(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except EnvironmentError as exc:
            msg = u'{exc}'.format(exc=exc)
            if sys.platform == 'win32':
                raise WindowsError(msg)
            else:
                raise OSError(msg)
    return res





import io
import importlib
import os
import fsx
from fstree import TYPE_FILE, TYPE_DIR

zipfile = importlib.import_module('zipfile')

class FakeZipFile(zipfile.ZipFile):
    def __init__(self, fsx_fake, *a, **kwa):
        zipfile.ZipFile.__init__(self, *a, **kwa)
        self._fsx_fake = fsx_fake

    def write(self, path): #, arcname, compress_type
        if self._fsx_fake._curdir:
            path_abs = self._fsx_fake._curdir + '/' + path
            nodes = self._fsx_fake.find(path_abs, TYPE_FILE)
            if nodes:
                path = path_abs
                filenode = nodes[0]
            else:
                filenode = self._fsx_fake._find_or_raise(path, TYPE_FILE)
        else:
            filenode = self._fsx_fake._find_or_raise(path, TYPE_FILE)
        filename = path

        #filenode = self._fsx_fake._find_or_raise(filename, TYPE_FILE)
        io_obj = filenode.get_exist_io(io.BytesIO)

        if self._fsx_fake._curdir:
            commonprefix = os.path.commonprefix([filename, self._fsx_fake._curdir])
            if filename[len(commonprefix)] == '/':
                zip_filepath = filename[len(commonprefix)+1:]
            else:
                zip_filepath = _cut_drive_or_root_prefix(filename)
        else:
            zip_filepath = _cut_drive_or_root_prefix(filename)
        self.writestr(zip_filepath, io_obj.getvalue())

def _cut_drive_or_root_prefix(path):
    res = path
    res = res.lstrip('/')
    if res[1:2] == ':':
        res = res[3:]
    return res

class Mixin(object):
    def _fake_zipfile_ZipFile(self, file, mode="r", compression=zipfile.ZIP_STORED, allowZip64=False):
        path = file
        if self._flip_backslashes:
            path = path.replace('\\', '/')

        if mode.startswith('r'):
            if self._curdir:
                path_abs = self._curdir + '/' + path
                nodes = self.find(path_abs, TYPE_FILE)
                if nodes:
                    path = path_abs
                    filenode = nodes[0]
                else:
                    filenode = self._find_or_raise(path, TYPE_FILE)
            else:
                filenode = self._find_or_raise(path, TYPE_FILE)

            io_obj = filenode.get_exist_io(io.BytesIO)
            io_obj.seek(0)

            res = FakeZipFile(self, io_obj, mode=mode, compression=compression, allowZip64=allowZip64)
            return res

        elif mode.startswith('a'):
            filenodes = self.find(path, TYPE_FILE)
            if not filenodes:
                # the file doesn't exist yet but the parent directory should.
                dirpath = os.path.dirname(path)
                self._find_or_raise(dirpath, TYPE_DIR)
                filenode = self.add_file(path)
            else:
                filenode = filenodes[0]

            io_obj = filenode.get_exist_io(io.BytesIO)

            res = FakeZipFile(self, io_obj, mode=mode, compression=compression, allowZip64=allowZip64)
            return res

        elif mode.startswith('w'):
            if path.startswith('/') or path[1:2] == ':':
                # HACK: We identify this as an absolute path.
                # We need tests for this.
                nodes = self.find(path, TYPE_FILE)
            elif self._curdir:
                path = self._curdir + '/' + path
                nodes = self.find(path, TYPE_FILE)
            else:
                nodes = self.find(path, TYPE_FILE)

            filenode = nodes[0] if nodes else self.add_file(path)
            io_obj = filenode.create_new_io(io.BytesIO)

            res = FakeZipFile(self, io_obj, mode=mode, compression=compression, allowZip64=allowZip64)
            return res

        else:
            msg = "Mode string must begin with one of 'r', 'w' or 'a', not '{}'.".format(mode)
            raise ValueError(msg)




import importlib
zipfile = importlib.import_module('zipfile')

__all__ = ["BadZipfile", "error", "ZIP_STORED", "ZIP_DEFLATED", "is_zipfile",
           "ZipInfo", "ZipFile", "PyZipFile", "LargeZipFile" ]

BadZipfile = zipfile.BadZipfile
error = zipfile.error
ZIP_STORED = zipfile.ZIP_STORED
ZIP_DEFLATED = zipfile.ZIP_DEFLATED
is_zipfile = zipfile.is_zipfile
ZipInfo = zipfile.ZipInfo
ZipFile = zipfile.ZipFile
PyZipFile = zipfile.PyZipFile
LargeZipFile = zipfile.LargeZipFile



import os
from fstree import TYPE_FILE, TYPE_DIR



class Mixin(object):
    def _fake_builtins_open(self, path, mode='r', encoding=None):
        if self._flip_backslashes:
            path = path.replace('\\', '/')

        if mode.startswith('r'):
            filenode = self._find_or_raise(path, TYPE_FILE)
            res = filenode.get_exist_io()
            res.seek(0)
            return res
        elif mode.startswith('a'):
            filenode = self._find_or_create_writable_filenode(path)
            res = filenode.get_exist_io()
            # Jump to end of stream so one can append to it.
            res.seek(0, 2)
            return res
        elif mode.startswith('w'):
            filenode = self._find_or_create_writable_filenode(path)
            res = filenode.create_new_io()
            return res
        else:
            msg = "Mode string must begin with one of 'r', 'w' or 'a', not '{}'.".format(mode)
            raise ValueError(msg)

    def _find_or_create_writable_filenode(self, path):
        filenodes = self.find(path, TYPE_FILE)
        if not filenodes:
            # the file doesn't exist yet but the parent directory should.
            dirpath = os.path.dirname(path)
            if dirpath == '':
                # there is no parent directory, just create the file
                filenode = self.add_file(path)
            else:
                # the parent directory should exist
                self._find_or_raise(dirpath, TYPE_DIR)
                filenode = self.add_file(path)
        else:
            filenode = filenodes[0]
        return filenode




import os
import ftplib
from fnmatch import fnmatch
from fstree import FsTree, TYPE_FILE, TYPE_DIR, node_matches_type
import fsx
import fsx.scandir


class Mixin(object):
    def _fake_ftplib_FTP(self, url):
        res = FakeFTP(self, url)
        return res

class FakeFTP(object):
    def __init__(self, fsx_fake, url):
        self._fsx_fake = fsx_fake
        self._url = 'ftp://'+url
        self._logged_in = False

    def storbinary(self, command, fp):
        if not self._logged_in:
            raise ftplib.error_perm('530 Please log in with USER and PASS first.')

        parts = command.split(' ', 1)
        if len(parts) == 2 and parts[0] == 'STOR':
            dst_path = _get_abs_url_path(self._url, parts[1])
            try:
                content = fp.read()
                fsx.open(dst_path, 'wb').write(content)
            except EnvironmentError as exc:
                msg = '{exc.__class__.__name__}: {exc}'.format(exc=exc)
                raise ftplib.error_perm(msg)
        else:
            raise NotImplementedError("Commands other than 'STOR' are currently not implemented.")

    def mkd(self, pathname):
        self._fsx_fake.add_dir(self._url + '/' + pathname)

    def nlst(self, argument):
        if not self._logged_in:
            raise ftplib.error_perm('530 Please log in with USER and PASS first.')

        res = []

        path = _get_abs_url_path(self._url, argument)
        for entry in fsx.scandir.scandir(path):
            res.append(entry.name)

        return res

    def retrbinary(self, argument, callback):
        if not self._logged_in:
            raise ftplib.error_perm('530 Please log in with USER and PASS first.')

        parts = argument.split(' ', 1)
        if len(parts) == 2 and parts[0] == 'RETR':
            path = _get_abs_url_path(self._url, parts[1])
            try:
                with fsx.open(path, 'rb') as file_:
                    callback(file_.read())
            except EnvironmentError as exc:
                msg = '{exc.__class__.__name__}: {exc}'.format(exc=exc)
                raise ftplib.error_perm(msg)
        else:
            raise NotImplementedError("Commands other than 'RETR' are currently not implemented.")

    def login(self, user, password):
        self._logged_in = True

    def prot_p(self):
        if not self._logged_in:
            raise AttributeError("FTP instance has no attribute 'prot_p'")

    def quit(self):
        self._logged_in = False

    def close(self):
        self._logged_in = False

def _get_abs_url_path(url, path):
    path = path.lstrip('/')
    path = url+'/'+path
    path = path.rstrip('/')
    return path



import os
from fnmatch import fnmatch
from fstree import FsTree, TYPE_FILE, TYPE_DIR, node_matches_type
import fsx


class Mixin(object):
    def _fake_glob_glob(self, pattern):
        if self._flip_backslashes:
            pattern = pattern.replace('\\', '/')

        res = list(self._fake_glob_iglob(pattern))
        return res

    def _fake_glob_iglob(self, pattern):
        if self._flip_backslashes:
            pattern = pattern.replace('\\', '/')

        pat_parts = pattern.split('/')
        res = _getMatchingNodePaths(self, pat_parts, 0)
        return res

def _getMatchingNodePaths(node, pat_parts, level):
    res = []
    islast = level == len(pat_parts) - 1
    for childnode in node.children:
        matches = fnmatch(childnode.name, pat_parts[level])
        if matches and islast:
            res.append(childnode.get_fspath())
        elif matches:
            res += _getMatchingNodePaths(childnode, pat_parts, level+1)
    return res




import importlib
os = importlib.import_module('os')

__all__ = ['chdir', 'getcwd', 'listdir', 'makedirs', 'rmdir', 'path', 'walk']

chdir = os.chdir
listdir = os.listdir
makedirs = os.makedirs
rmdir = os.rmdir
getcwd = os.getcwd
remove = os.remove

def walk(top):
    # TODO: Not yet supported parameters: topdown=True, onerror=None, followlinks=False
    try:
        walker = importlib.import_module('scandir')
    except:
        walker = os
    res = walker.walk(top)
    return res


class path:
    @staticmethod
    def isfile(path):
        res = os.path.isfile(path)
        return res

    @staticmethod
    def isdir(path):
        res = os.path.isdir(path)
        return res

    @staticmethod
    def exists(path):
        res = os.path.exists(path)
        return res

    @staticmethod
    def abspath(path):
        res = os.path.abspath(path)
        return res



import os
import scandir
from fstree import TYPE_FILE, TYPE_ALL, TYPE_DIR, TYPE_SYMLINK, node_matches_type


__all__ = ['Mixin']


class Mixin(object):
    def _fake_scandir_scandir(self, path):
        parent_node = self._find_or_raise(path, TYPE_ALL)
        for child_node in sorted(parent_node.children):
            dir_entry = FakeDirEntry(child_node)
            yield dir_entry


class FakeDirEntry(scandir.GenericDirEntry):
    __slots__ = ('name', '_stat', '_lstat', '_scandir_path', '_path', '_node')

    def __init__(self, node):
        self._node = node
        dirname, basename = os.path.split(node.get_fspath())
        scandir.GenericDirEntry.__init__(self, dirname, basename)

    def stat(self, follow_symlinks=True):
        if follow_symlinks:
            if self._stat is None:
                self._stat = _fake_stat(self._node)
            return self._stat
        else:
            if self._lstat is None:
                self._lstat = _fake_lstat(self.path)
            return self._lstat

def _fake_stat(node):
    return Stat(node)

def _fake_lstat(node):
    raise NotImplementedError()

class Stat(object):
    def __init__(self, node):
        self._node = node

        self.st_mode = self._get_st_mode()

        self.st_ino = property(self._not_implemented)  # inode number,
        self.st_dev = property(self._not_implemented)  # device,
        self.st_nlink = property(self._not_implemented)  # number of hard links,
        self.st_uid = property(self._not_implemented)  # user id of owner,
        self.st_gid = property(self._not_implemented)  # group id of owner,
        self.st_size = property(self._not_implemented) # size of file, in bytes,
        self.st_atime = property(self._not_implemented)  # time of most recent access,
        self.st_mtime = property(self._not_implemented)  # time of most recent content modification,
        self.st_ctime = property(self._not_implemented)  # platform dependent; time of most recent metadata change on Unix, or the time of creation on Windows)

    def _not_implemented(self):
        raise NotImplementedError

    def _get_st_mode(self):
        if node_matches_type(self._node, TYPE_DIR):
            return 0o40777
        elif node_matches_type(self._node, TYPE_FILE):
            return 0o100666
        elif node_matches_type(self._node, TYPE_SYMLINK):
            return 0o40777L
        else:
            raise ValueError('Unknown file type: {}'.format(self._node.get_fspath()))



import sys
from fstree import TYPE_FILE, TYPE_DIR
from ._helpers import env_error_to_os_specific

import importlib
os = importlib.import_module('os')

def _is_win_abspath(path):
    res = sys.platform == 'win32' and path[1:2] == ':'
    return res

def _is_win_driveletter_with_trail_slash(path):
    return len(path) == 3 and path[1] == ':'

class Mixin(object):
    _curdir = ''

    def _fake_os_chdir(self, path):
        if self._flip_backslashes:
            path = path.replace('\\', '/')

        if path == '/':
            path = ''

        dirnode = self._find_or_raise(path, TYPE_DIR)

        if _is_win_driveletter_with_trail_slash(path):
            # self._setCurDir(path[:-1])
            self._curdir = path[:-1]
        else:
            # self._setCurDir(path.rstrip('/'))
            self._curdir = path.rstrip('/')

    def _fake_os_getcwd(self):
        if len(self._curdir) == 2 and self._curdir[1] == ':':
            return self._curdir + '/'
        else:
            return self._curdir

    def _get_abspath_from_relative(self, path):
        if self._curdir == '':
            res = path.lstrip('/')
            return res
        else:
            res = self._curdir + '/' + path.lstrip('/')
            return res

    def _fake_os_path_exists(self, path):
        if path == '':
            return False

        if self._flip_backslashes:
            path = path.replace('\\', '/')

        abspath_from_relative = self._get_abspath_from_relative(path)
        found_as_relative = bool(self.find(abspath_from_relative))
        if found_as_relative:
            return True
        else:
            res = bool(self.find(path))
            return res

    def _fake_os_path_isfile(self, path):
        if self._flip_backslashes:
            path = path.replace('\\', '/')
        nodes = self.find(path, TYPE_FILE)
        if nodes:
            return True
        return False

    def _fake_os_path_isdir(self, path):
        if self._flip_backslashes:
            path = path.replace('\\', '/')
        nodes = self.find(path, TYPE_DIR)
        if nodes:
            return True
        return False

    def _fake_os_listdir(self, path):
        if self._flip_backslashes:
            path = path.replace('\\', '/')

        res = []
        path = path.rstrip('/')
        dirnode = self._find_or_raise(path, TYPE_DIR)
        res = [child.name for child in dirnode.children]
        return res

    def _fake_os_walk(self, top):
        # Not yet supported parameters: topdown=True, onerror=None, followlinks=False

        if self._flip_backslashes:
            top = top.replace('\\', '/')
        node = self._find_or_raise(top, TYPE_DIR)

        if top.count('/'):
            prefix = top.rsplit('/', 1)[0] + '/'
        else:
            prefix = ''

        for root, dirname, filenames in node.walk():
            if root is None:
                root = top
            root = prefix + root
            yield root, dirname, filenames

    def _fake_os_makedirs(self, name):
        if self._fake_os_path_exists(name):
            if sys.platform == 'win32':
                raise WindowsError(183, "Cannot create a file when that file already exists: '{}'".format(name))
            else:
                raise OSError(17, "File exists: '{}'".format(name))

        name = os.path.normpath(name).replace('\\', '/')
        if name[:1].isalnum() and self._curdir:
            # make an absolute path using the current directory
            name = self._curdir + '/' + name

        self.add_dir(name)

    def _fake_os_rmdir(self, path):
        dirnode = self._find_or_raise(path, TYPE_DIR)
        if dirnode.children:
            if sys.platform == 'win32':
                raise WindowsError(145, "The directory is not empty: '{}'".format(path))
            else:
                raise OSError(66, "Directory not empty: '{}'".format(path))
        dirnode.remove()

    def _fake_os_remove(self, path):
        filenode = self._find_or_raise(path, TYPE_FILE)
        filenode.remove()




import codecs
import importlib

# Exports for backwards-compatibility (please use `fsx.os` instead):
from .os import path, listdir, makedirs, rmdir, walk

isfile = path.isfile
isdir = path.isdir
exists = path.exists

__all__ = [
    # Exports for backwards-compatibility (please use `fsx.os` instead):
    'isfile', 'isdir', 'exists',
    'listdir', 'makedirs', 'rmdir', 'walk'

    # Builtins:
    'open',
]

def open(filename, mode='r', encoding=None):
    res = codecs.open(filename, mode, encoding=encoding)
    return res



from fstree import FsTree
import fsx
import fsx.zipfile
import fsx.tempfile
import fsx.scandir
import fsx.glob
import fsx.ftplib
from . import _fake_os
from . import _fake_builtins
from . import _fake_glob
from . import _fake_tempfile
from . import _fake_zipfile
from . import _fake_scandir
from . import _fake_ftplib
from . _helpers import env_error_to_os_specific


class FsxFakeFsTree(FsTree,
                    _fake_builtins.Mixin,
                    _fake_os.Mixin,
                    _fake_glob.Mixin,
                    _fake_tempfile.Mixin,
                    _fake_zipfile.Mixin,
                    _fake_scandir.Mixin,
                    _fake_ftplib.Mixin,
    ):
    ''' A virtual file system tree supporting some common file system functions. '''

    def __init__(self, monkeypatch, flip_backslashes=None):
        FsTree.__init__(self, flip_backslashes=flip_backslashes)

        monkeypatch.setattr(fsx,          'open',               self._fake_builtins_open)
        monkeypatch.setattr(fsx.glob,     'iglob',              self._fake_glob_iglob)
        monkeypatch.setattr(fsx.glob,     'glob',               self._fake_glob_glob)
        monkeypatch.setattr(fsx.tempfile, 'NamedTemporaryFile', self._fake_tempfile_NamedTemporaryFile)
        monkeypatch.setattr(fsx.zipfile,  'ZipFile',            self._fake_zipfile_ZipFile)
        monkeypatch.setattr(fsx.scandir,  'scandir',            self._fake_scandir_scandir)
        monkeypatch.setattr(fsx.ftplib,   'FTP',                self._fake_ftplib_FTP)

        monkeypatch.setattr(fsx.os,       'listdir',            self._fake_os_listdir)
        monkeypatch.setattr(fsx.os,       'makedirs',           self._fake_os_makedirs)
        monkeypatch.setattr(fsx.os,       'rmdir',              self._fake_os_rmdir)
        monkeypatch.setattr(fsx.os,       'walk',               self._fake_os_walk)
        monkeypatch.setattr(fsx.os.path,  'exists',             self._fake_os_path_exists)
        monkeypatch.setattr(fsx.os.path,  'isfile',             self._fake_os_path_isfile)
        monkeypatch.setattr(fsx.os.path,  'isdir',              self._fake_os_path_isdir)
        monkeypatch.setattr(fsx.os,       'chdir',              self._fake_os_chdir)
        monkeypatch.setattr(fsx.os,       'getcwd',             self._fake_os_getcwd)
        monkeypatch.setattr(fsx.os,       'remove',             self._fake_os_remove)

        # Legacy root-namespace functions, do not add new ones.
        monkeypatch.setattr(fsx,          'listdir',            self._fake_os_listdir)
        monkeypatch.setattr(fsx,          'makedirs',           self._fake_os_makedirs)
        monkeypatch.setattr(fsx,          'rmdir',              self._fake_os_rmdir)
        monkeypatch.setattr(fsx,          'walk',               self._fake_os_walk)
        monkeypatch.setattr(fsx,          'exists',             self._fake_os_path_exists)
        monkeypatch.setattr(fsx,          'isfile',             self._fake_os_path_isfile)
        monkeypatch.setattr(fsx,          'isdir',              self._fake_os_path_isdir)

    @env_error_to_os_specific
    def _find_or_raise(self, path, type_):
        return FsTree._find_or_raise(self, path, type_)

    @env_error_to_os_specific
    def _raise_not_found(self, file_path):
        return FsTree._raise_not_found(self, file_path)




import os
import weakref
import errno
import io
import importlib
from fstree import TYPE_DIR
from fstree._FileNode import FileNode, FileStringIO
import fsx

tempfile = importlib.import_module('tempfile')


class Mixin(object):
    def _fake_tempfile_NamedTemporaryFile(self, mode='w+b', suffix="", prefix=tempfile.template, dir=None, delete=True):
        if dir is None:
            dir = tempfile.gettempdir()

        path = _get_new_path(dir, prefix, suffix)
        if self._flip_backslashes:
            path = path.replace('\\', '/')

        if mode.startswith('w') or mode.startswith('a'):
            filenode = _add_named_node(self, path, delete)
            res = filenode.create_new_io()
            return res

        else:
            msg = "Mode string must begin with one of 'r', 'w' or 'a', not '{}'.".format(mode)
            raise ValueError(msg)


def _add_named_node(fstree, path, delete):
    dirname, basename = os.path.split(path)
    nodes = fstree.find(dirname, TYPE_DIR)
    dirnode = nodes[0] if nodes else fstree.add_dir(dirname)
    res = NamedTemporaryFileNode(path, delete, dirnode)
    return res

def _get_new_path(dir, prefix, suffix):
    ''' Extracted from `tempfile.py` to build proper temp-paths. '''
    names = tempfile._get_candidate_names()

    for seq in xrange(tempfile.TMP_MAX):
        name = names.next()
        res = os.path.join(dir, prefix + name + suffix)
        if fsx.exists(res):
            continue
        return res

    raise IOError(errno.EEXIST, "No usable temporary file name found")

class AutoDeletableFileStringIO(FileStringIO):
    def __init__(self, filenode, path, delete):
        self._filenode = filenode
        self._delete_on_close = delete
        self.name = path
        FileStringIO.__init__(self)

    def close(self):
        if self._delete_on_close:
            self._filenode.remove()
        else:
            value = self.getvalue()
            parent_node = self._filenode.parent
            self._filenode.remove()

            new_filenode = FileNode(
                os.path.basename(self.name),
                parent_node,
            )
            new_filenode.get_exist_io().write(value)

        FileStringIO.close(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_info, exc_tb):
        self.seek(0)
        self.close()
        return False

class NamedTemporaryFileNode(FileNode):
    '''
        Custom `FileNode` class holding a weak reference
        to the IO object.
    '''
    def __init__(self, filepath, delete, parent):
        name = os.path.basename(filepath)
        FileNode.__init__(self, name, parent)
        self._filepath = filepath
        self._delete_on_close = delete
        self._io_ref = None

    def create_new_io(self):
        io = AutoDeletableFileStringIO(
            self,
            self._filepath,
            self._delete_on_close
        )

        # Use a weakref so the file-descriptor can be
        # properly garbage-collected like if the scope changes.
        def cb(ref):
            if ref() is None:
                # there are no more references
                if self._delete_on_close:
                    self.remove()

        self._io_ref = weakref.ref(io, cb)
        return io

    def get_exist_io(self):
        if self._io_ref is None:
            # An io object wasn't created yet.
            self.create_new_io()
            io = self._io_ref()
            return io
        else:
            io = self._io_ref()
            return io

